(*
 * Extract all the definitions from the given Mizar article.
 * Date: 2006-10-29
 * Copyright (C) 2006 Michael Nedzelsky (MichaelNedzelsky@yandex.ru)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Mizast

let mizar_dir = Sys.getenv("MIZFILES");;
let miz_article_path name = Filename.concat (Filename.concat mizar_dir "mml") (name ^ ".miz");;
let mml_vct = Filename.concat mizar_dir "mml.vct"
let voc_dir = ref (Sys.getcwd())

let vct_map = Voc.read_vct_file mml_vct

let rec local_print_node node =
  let rec print_list tk_start tk_end node_list =
    match node_list with
    | [] -> print_string (tokens_to_txt tk_start tk_end)
    | h :: t -> let () = print_string (tokens_to_txt tk_start h.nt_first_tk) in
                let () = local_print_node h in
                print_list h.nt_last_tk tk_end t
  in
  match node.nt_type with
  | Justification -> if node.nt_first_tk != node.nt_last_tk then print_string "<Justification>"
  | _ -> print_list node.nt_first_tk node.nt_last_tk node.nt_sublist

let rec print_defs node =
  match node.nt_type with
  | Structure_definition
  | Mode_definition
  | Functor_definition
  | Attribute_definition -> (local_print_node node; print_endline "")
  | _ -> if node.nt_expanded then List.iter print_defs node.nt_sublist

let one_file voc_dir miz_filename =
  begin
    try
      let ast_root = Mizast.miz_file_to_ast vct_map voc_dir miz_filename in
      print_defs ast_root
    with Lexer.Token_error (line, pos) ->
      prerr_endline (miz_filename ^ ": Token error at line " ^ (string_of_int line) ^ ", pos " ^ (string_of_int pos))
  end

let arg_list = [("-vocdir", Arg.String (function s -> voc_dir := s), "path to the dict directory")]

let f = one_file !voc_dir

let usage = "Usage: extract_defs [-vocdir dict] text/test.miz > out.txt"

let () = Arg.parse arg_list f usage
