(*
 * Usefull functions for tokenizing Mizar articles.
 * Date: 2006-10-15
 * Copyright (C) 2006 Michael Nedzelsky (MichaelNedzelsky@yandex.ru)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Voc

let local_input_line in_channel =
  let line = input_line in_channel in
  let len = String.length line in
  if len > 0 && line.[len-1] = '\r' then String.sub line 0 (len-1) else line

let is_ext_letter ch = match ch with 'a'..'z' | 'A'..'Z' | '_' -> true | _ -> false
let is_letter_or_digit ch = match ch with 'a'..'z' | 'A'..'Z' | '0'..'9' -> true | _ -> false
let is_white ch = match ch with '\t' | ' ' -> true | _ -> false

(* Result is true or false *)
let is_numeral s =
  let len = String.length s in
  let rec is_numeral1 pos =
    if pos < len then
      match s.[pos] with
      | '0' when pos=0 -> false
      | '0'..'9' -> is_numeral1 (pos+1)
      | _ -> false
    else true
  in (is_numeral1 0) && (len > 0)

(* Identifier is a string of letters, digits, underscores (  _  ) and apostrophes (  '  ), that is neither a reserved word nor symbol nor numeral. *)
let is_ident s =
  let len = String.length s in
  let rec is_id pos =
    if pos < len then match s.[pos] with
      'a'..'z'| 'A'..'Z'| '_' | '\'' | '0'..'9' -> is_id (pos+1)
    | _ -> false
    else true
  in (is_id 0) && (len > 0)
  

(* The result is the last position pos1 such that s.[pos], ..., s.[pos1-1] are letters or digits. *)
let rec ltr_dgt_seq s pos = match s.[pos] with
  ch when is_letter_or_digit ch -> ltr_dgt_seq s (pos+1)
  | _ -> pos

(* The result is the last position pos1 such that s.[pos], ..., s.[pos1-1] forms an identifier. *)
let ident s pos = if is_ext_letter s.[pos] then ltr_dgt_seq s (pos+1) else pos

(* The result is the last position pos1 such that s.[pos], ..., s.[pos1-1] forms a whitespace. *)
let rec white_space s pos = if is_white s.[pos] then white_space s (pos+1) else pos

type t_miz_token = Miz_according      | Miz_aggregate     | Miz_and            | Miz_antonym
                  | Miz_as            | Miz_associativity | Miz_assume         | Miz_asymmetry
		  | Miz_attr          | Miz_be            | Miz_begin          | Miz_being
                  | Miz_by            | Miz_canceled      | Miz_case           | Miz_cases
                  | Miz_cluster       | Miz_coherence     | Miz_commutativity  | Miz_compatibility
                  | Miz_connectedness | Miz_consider      | Miz_consistency    | Miz_constructors
                  | Miz_contradiction | Miz_correctness   | Miz_def            | Miz_deffunc
                  | Miz_define        | Miz_definition    | Miz_definitions    | Miz_defpred
                  | Miz_end           | Miz_environ       | Miz_equals         | Miz_ex
                  | Miz_exactly       | Miz_existence     | Miz_for            | Miz_from
                  | Miz_func          | Miz_given         | Miz_hence          | Miz_hereby
                  | Miz_holds         | Miz_idempotence   | Miz_identify       | Miz_if
                  | Miz_iff           | Miz_implies       | Miz_involutiveness | Miz_irreflexivity
                  | Miz_is            | Miz_it            | Miz_let            | Miz_means
                  | Miz_mode          | Miz_non           | Miz_not            | Miz_notation
                  | Miz_notations     | Miz_now           | Miz_of             | Miz_or
                  | Miz_otherwise     | Miz_over          | Miz_per            | Miz_pred
                  | Miz_prefix        | Miz_projectivity  | Miz_proof          | Miz_at_proof
                  | Miz_provided      | Miz_qua           | Miz_reconsider     | Miz_redefine
                  | Miz_reflexivity   | Miz_registration  | Miz_registrations  | Miz_requirements
                  | Miz_reserve       | Miz_sch           | Miz_scheme         | Miz_schemes
                  | Miz_section       | Miz_selector      | Miz_set            | Miz_st
                  | Miz_struct        | Miz_such          | Miz_suppose        | Miz_symmetry
                  | Miz_synonym       | Miz_take          | Miz_that           | Miz_the
                  | Miz_then          | Miz_theorem       | Miz_theorems       | Miz_thesis
                  | Miz_thus          | Miz_to            | Miz_transitivity   | Miz_uniqueness
                  | Miz_vocabularies  | Miz_when          | Miz_where          | Miz_with
                  | Miz_wrt
		  (* special symbols *)
		  |  Miz_comma    (*  ','   *)
		  | Miz_semicolon (*  ';'   *)
		  | Miz_colon     (*  ':'   *)
		  | Miz_lp        (*  '('   *)
		  | Miz_rp        (*  ')'   *)
		  | Miz_lb        (*  '['   *)
		  | Miz_rb        (*  ']'   *)
		  | Miz_lc        (*  '{'   *)
		  | Miz_rc        (*  '}'   *)
		  | Miz_eq        (*  '='   *)
		  | Miz_amp       (*  '&'   *)
		  | Miz_imp       (*  '->'  *)
		  | Miz_seq_eq    (*  '.='  *)
                  | Miz_dots      (*  '...' *)
		  | Miz_S1        (*  '$1'  *)
		  | Miz_S2        (*  '$2'  *)
		  | Miz_S3        (*  '$3'  *)
		  | Miz_S4        (*  '$4'  *)
		  | Miz_S5        (*  '$5'  *)
		  | Miz_S6        (*  '$6'  *)
		  | Miz_S7        (*  '$7'  *)
		  | Miz_S8        (*  '$8'  *)
		  | Miz_S9        (*  '$9'  *)
		  | Miz_S10       (*  '$10' *)
		  | Miz_lpound    (*  '(#'  *)
		  | Miz_rpound    (*  '#)'  *)
                  | Miz_tk_numeral    (* Numeral constant *)
		  | Miz_tk_ident      (* Identifier *)
		  | Miz_tk_symbol of Voc.symbol_qualifier (* Symbol from additional lexicon *)
		  | Miz_tk_whitespace (* Whitespace *)
		  | Miz_tk_comment    (* :: <Text> *)
                  | Miz_tk_eol        (* End of line *)
                  | Miz_tk_empty      (* At the beginning of the text *)
                  | Miz_tk_eof        (* At the end of the text *)


let miz_kw_list =
  [ ("according", Miz_according); ("aggregate", Miz_aggregate); ("and", Miz_and); ("antonym", Miz_antonym);
    ("as", Miz_as); ("associativity", Miz_associativity); ("assume", Miz_assume); ("asymmetry", Miz_asymmetry);
    ("attr", Miz_attr); ("be", Miz_be); ("begin", Miz_begin); ("being", Miz_being);
    ("by", Miz_by); ("canceled", Miz_canceled); ("case", Miz_case); ("cases", Miz_cases);
    ("cluster", Miz_cluster); ("coherence", Miz_coherence); ("commutativity", Miz_commutativity); ("compatibility", Miz_compatibility);
    ("connectedness", Miz_connectedness); ("consider",    Miz_consider); ("consistency",  Miz_consistency); ("constructors",  Miz_constructors);
    ("contradiction", Miz_contradiction); ("correctness", Miz_correctness); ("def", Miz_def); ("deffunc", Miz_deffunc);
    ("define", Miz_define); ("definition", Miz_definition); ("definitions", Miz_definitions); ("defpred", Miz_defpred);
    ("end", Miz_end); ("environ", Miz_environ); ("equals", Miz_equals); ("ex", Miz_ex);
    ("exactly", Miz_exactly); ("existence", Miz_existence); ("for", Miz_for); ("from", Miz_from);
    ("func", Miz_func); ("given", Miz_given); ("hence", Miz_hence); ("hereby", Miz_hereby);
    ("holds", Miz_holds); ("idempotence", Miz_idempotence); ("identify", Miz_identify); ("if", Miz_if);
    ("iff",         Miz_iff); ("implies", Miz_implies); ("involutiveness", Miz_involutiveness); ("irreflexivity", Miz_irreflexivity);
    ("is", Miz_is); ("it", Miz_it); ("let", Miz_let); ("means", Miz_means);
    ("mode", Miz_mode); ("non", Miz_non); ("not", Miz_not); ("notation", Miz_notation);
    ("notations", Miz_notations); ("now", Miz_now); ("of", Miz_of); ("or", Miz_or);
    ("otherwise", Miz_otherwise); ("over", Miz_over); ("per", Miz_per); ("pred", Miz_pred);
    ("prefix", Miz_prefix); ("projectivity", Miz_projectivity); ("proof", Miz_proof); ("@proof", Miz_at_proof);
    ("provided", Miz_provided); ("qua", Miz_qua); ("reconsider", Miz_reconsider); ("redefine", Miz_redefine);
    ("reflexivity", Miz_reflexivity); ("registration", Miz_registration); ("registrations", Miz_registrations); ("requirements", Miz_requirements);
    ("reserve", Miz_reserve); ("sch", Miz_sch); ("scheme", Miz_scheme); ("schemes", Miz_schemes);
    ("section", Miz_section); ("selector", Miz_selector); ("set", Miz_set); ("st", Miz_st);
    ("struct", Miz_struct); ("such", Miz_such); ("suppose", Miz_suppose); ("symmetry", Miz_symmetry);
    ("synonym", Miz_synonym); ("take", Miz_take); ("that", Miz_that); ("the", Miz_the);
    ("then", Miz_then); ("theorem", Miz_theorem); ("theorems", Miz_theorems); ("thesis", Miz_thesis);
    ("thus", Miz_thus); ("to", Miz_to); ("transitivity", Miz_transitivity); ("uniqueness", Miz_uniqueness);
    ("vocabularies",  Miz_vocabularies); ("when", Miz_when); ("where", Miz_where); ("with", Miz_with); ("wrt", Miz_wrt);
    (* special symbols *)
    (",", Miz_comma); (";", Miz_semicolon); (":", Miz_colon); ("(", Miz_lp); (")", Miz_rp); ("[", Miz_lb); ("]", Miz_rb); ("{", Miz_lc); ("}", Miz_rc);
    ("=", Miz_eq); ("&", Miz_amp); ("->", Miz_imp); (".=", Miz_seq_eq); ("...", Miz_dots);
    ("$1", Miz_S1); ("$2", Miz_S2); ("$3", Miz_S3); ("$4", Miz_S4); ("$5", Miz_S5);
    ("$6", Miz_S6); ("$7", Miz_S7); ("$8", Miz_S8); ("$9", Miz_S9); ("$10", Miz_S10); ("(#", Miz_lpound); ("#)", Miz_rpound)
  ]


(* module MapOfString = Map.Make(String) *)

module MizToken =
  struct
    type t = t_miz_token
    let compare v1 v2 = if v1 < v2 then -1 else if v1 > v2 then 1 else 0
  end

module MapOfMiz = Map.Make(MizToken)

let rec add_list m l = 
  let add_pair h = match h with
    (s,miz) -> MapOfString.add s miz m
  in
  match l with
    [] -> m
    | h :: t -> add_list (add_pair h) t

let miz_map = add_list MapOfString.empty miz_kw_list

type t_token = {
    tk_type: t_miz_token;
    tk_line: int;
    tk_pos: int;
    tk_len: int;
    tk_text: string;
    tk_prev: t_token;
    mutable tk_next: t_token option }

let create_token t line pos text tk_prev = {
  tk_type = t;
  tk_line = line;
  tk_pos = pos;
  tk_len = String.length text;
  tk_text = String.copy text;
  tk_prev = tk_prev;
  tk_next = None }

let rec empty_token = {
  tk_type = Miz_tk_empty;
  tk_line = 0;
  tk_pos = 0;
  tk_len = 0;
  tk_text = "";
  tk_prev = empty_token;
  tk_next = None }

let create_eol_token line pos tk_prev = {
  tk_type = Miz_tk_eol;
  tk_line = line;
  tk_pos = pos;
  tk_len = 0;
  tk_text = "";
  tk_prev = tk_prev;
  tk_next = None }

let create_eof_token line pos tk_prev = {
  tk_type = Miz_tk_eof;
  tk_line = line;
  tk_pos = pos;
  tk_len = 0;
  tk_text = "";
  tk_prev = tk_prev;
  tk_next = None }

type t_state =
  { st_in: in_channel;
    mutable st_line_no: int;
    mutable st_line: string;
    mutable st_token: t_token;
    mutable st_next: int;     (*   st_next <= st_next_ws <= st_len <= String.length st_line *) 
    mutable st_next_ws: int;
    mutable st_len: int;      (*   st_line_cmt = true <=> st_len < String.length st_line    *)
    mutable st_voc: Voc.t_voc;
    mutable st_line_cmt: bool;
    mutable st_eol: bool }

let print_state st =
  begin
    print_string ("state(line_no=" ^ (string_of_int st.st_line_no));
    print_string (" next=" ^ (string_of_int st.st_next));
    print_string (" next_ws=" ^ (string_of_int st.st_next_ws));
    print_string (" len=" ^ (string_of_int st.st_len));
    print_string (" text='" ^ st.st_token.tk_text ^ "'");
    print_string (if st.st_line_cmt then " cmt" else " non-cmt");
    print_string (if st.st_eol then " eol" else " non-eol");
    print_string "\n"
  end

(* Given string s must not contain whitespaces *)
let determine_token_type s voc =
  try MapOfString.find s miz_map
  with Not_found -> if is_numeral s then Miz_tk_numeral else try Miz_tk_symbol (MapOfString.find s voc) with Not_found -> if is_ident s then Miz_tk_ident else raise Not_found

let rec verify_cmt s pos =
  let len = String.length s in
    if pos+1 >= len then (false, len)
    else if s.[pos] = ':' && s.[pos+1] = ':' then (true, pos)
      else verify_cmt s (pos+1)
  
let rec find_ws s pos len = if pos >= len then len else if is_white s.[pos] then pos else find_ws s (pos+1) len
let rec find_non_ws s pos len = if pos >= len then len else if not (is_white s.[pos]) then pos else find_non_ws s (pos+1) len

let create_state in_ch =
  let s = local_input_line in_ch in
  let (cmt, len) = verify_cmt s 0 in
  let ws = find_ws s 0 len in
    { st_in = in_ch; st_line_no=1; st_token = empty_token; st_line = s; st_next = 0; st_next_ws = ws; st_len = len; st_voc = MapOfString.empty; st_line_cmt = cmt; st_eol = true }


let state_next state =
  let in_ch = state.st_in in
  let s = local_input_line in_ch in
  let (cmt, len) = verify_cmt s 0 in
  let ws = find_ws s 0 len in
    ( state.st_line_no <- state.st_line_no + 1; state.st_line <- s; state.st_next <- 0; state.st_next_ws <- ws; state.st_len <- len; state.st_line_cmt <- cmt; state.st_eol <- true )

let close_state state = close_in state.st_in

(*  Result is the maximal prefix of s[pos1], ..., s[pos2-1] which forms a valid Mizar token. *)
let rec read_non_ws_token s pos1  pos2 voc = let len = pos2 - pos1 in let ()= if len = 0 then raise Not_found in let s1 = String.sub s pos1 len in
  try (determine_token_type s1 voc, pos2)  with Not_found -> read_non_ws_token s pos1 (pos2-1) voc

exception Token_error of int * int

let rec read_token state =
  let read_main () =
    let (miz_type, next_pos) = try read_non_ws_token state.st_line state.st_next state.st_next_ws state.st_voc with Not_found -> raise (Token_error (state.st_line_no, state.st_next)) in
    let tk = create_token miz_type state.st_line_no state.st_next (String.sub state.st_line state.st_next (next_pos - state.st_next)) state.st_token in
    (state.st_next <- next_pos; state.st_token.tk_next <- Some tk;  state.st_token <- tk; tk) in
  let read_eol () =
    let tk = create_eol_token state.st_line_no state.st_next state.st_token in
    (state.st_eol <- false; state.st_token.tk_next <- Some tk; state.st_token <- tk; tk) in
  let read_eof () =
    let tk = create_eof_token state.st_line_no state.st_next state.st_token in
    (state.st_eol <- false; state.st_token.tk_next <- Some tk; state.st_token <- tk; tk) in
  let read_cmt ()  =
    let tk =
      let len1 = String.length state.st_line in
      create_token Miz_tk_comment state.st_line_no state.st_next (String.sub state.st_line state.st_next (len1 - state.st_next)) state.st_token in
      (state.st_line_cmt <- false; state.st_token.tk_next <- Some tk; state.st_token <- tk; tk) in
  let read_white () =
    let next_non_ws = find_non_ws state.st_line state.st_next state.st_len in
    let ws = find_ws state.st_line next_non_ws state.st_len in
    let len = next_non_ws - state.st_next in
    let tk = create_token Miz_tk_whitespace state.st_line_no state.st_next (String.sub state.st_line state.st_next len) state.st_token in
      (state.st_next <- next_non_ws; state.st_next_ws <- ws; state.st_token.tk_next <- Some tk; state.st_token <- tk; tk) in
  if state.st_next < state.st_next_ws then read_main() else
  if state.st_next_ws < state.st_len then read_white() else
  if state.st_line_cmt then read_cmt() else
  if state.st_eol then read_eol() else try (state_next state; read_token state) with End_of_file -> read_eof()

let rec read_main_token state =
  let tk = read_token state in
    match tk.tk_type with
    | Miz_tk_whitespace | Miz_tk_comment | Miz_tk_eol  -> read_main_token state
    | _  -> tk

let print_token tk = Printf.printf "token line=%d pos=%d text=%s\n" tk.tk_line tk.tk_pos tk.tk_text

type t_filename_list = t_token list
type lib_directive = Notations | Constructors | Registrations | Definitions | Theorems | Schemes

type ('a,'b) t_directive = Voc_directive of 'a | Lib_directive of 'a *'b | Req_directive of 'a

type t_miz_directive = (t_filename_list, lib_directive) t_directive

type t_environ = t_miz_directive list

(* filename (',' filename)* ';' *)
let rec read_filenames_list state =
    let get_token () = state.st_token in
    let res = ref [] in
    begin
      res := !res @ [get_token()]; ignore (read_main_token state);
      while state.st_token.tk_type = Miz_comma do
        ignore (read_main_token state);
        res := !res @ [get_token()]; ignore (read_main_token state);
      done;
      if state.st_token.tk_type <> Miz_semicolon then raise (Invalid_argument "expected ;") else ignore (read_main_token state);
      !res;
    end

let read_voc_directive state =
  let () = if state.st_token.tk_type <> Miz_vocabularies then raise (Invalid_argument "expected vocabularies") in
  let () = ignore (read_main_token state) in
  Voc_directive (read_filenames_list state)

let opt_read_directive state =
  let read_list () =
  begin
    ignore (read_main_token state);
    read_filenames_list state
  end in
  match state.st_token.tk_type with
    Miz_vocabularies  -> Voc_directive (read_list ())
  | Miz_requirements  -> Req_directive (read_list ())
  | Miz_notations     -> Lib_directive ((read_list ()), Notations)
  | Miz_constructors  -> Lib_directive ((read_list ()), Constructors)
  | Miz_registrations -> Lib_directive ((read_list ()), Registrations)
  | Miz_definitions   -> Lib_directive ((read_list ()), Definitions)
  | Miz_theorems      -> Lib_directive ((read_list ()), Theorems)
  | Miz_schemes       -> Lib_directive ((read_list ()), Schemes)
  | _ -> raise Not_found

let read_directives state =
  let rec read_list () = 
    try let d = opt_read_directive state in d :: (read_list()) with Not_found -> []
  in
  read_list()

let read_env_decl state =
  let () = if state.st_token.tk_type <> Miz_environ then raise (Invalid_argument "expected environ") in
  let () = ignore (read_main_token state) in
  read_directives state

let read_environ filename =
  let st = create_state (open_in filename) in
  let () = ignore (read_main_token st) in
  let res = read_env_decl st in
  (close_state st; res)

(* Returns a pair of tokens. *)
let read_env filename =
  let st = create_state (open_in filename) in
  let start_tk = st.st_token in
  let () = ignore (read_main_token st) in
  let () = ignore (read_env_decl st) in
  let first_tk = match start_tk.tk_next with Some x -> x | None -> failwith "expected first token" in
  (close_state st; (first_tk, st.st_token))

let get_voc_names env =
  let add res d =
    match d with
      Voc_directive ts -> res @ ts
    | _ -> res
  in
  List.fold_left add [] env

let add_voc_by_name v voc_name mv  =
  let v_x = MapOfString.find voc_name mv in
  let f key data v = MapOfString.add key data v in
  MapOfString.fold f v_x v

exception Vocabulary_not_found of string

let add_to_voc voc xs mv voc_dir  =
  let add v x =
    let text = x.tk_text in
    let v_x = try MapOfString.find text mv with Not_found ->
      try Voc.read_vocabulary voc_dir text with Not_found -> raise (Vocabulary_not_found text) in
    let f key data v = MapOfString.add key data v in
    MapOfString.fold f v_x v
  in
  List.fold_left add voc xs

(* Returns the pair (first token of article, first token of Text-Proper). *)
let read_tokens filename mv voc_dir =
  let st = create_state (open_in filename) in
  let read_to_end () =
    let rec read_inner tk =
      if tk.tk_type <> Miz_tk_eof then read_inner (read_main_token st) in
    read_inner st.st_token in
  let start_tk = st.st_token in
  let () = ignore (read_main_token st) in
  let env = read_env_decl st in
  let first_tk = match start_tk.tk_next with Some x -> x | None -> failwith "expected first token" in
  let res_tk = st.st_token in
  let voc_names = get_voc_names env in
  let voc = add_voc_by_name MapOfString.empty "HIDDEN" mv  in
  let voc = add_to_voc voc voc_names mv voc_dir in
  let () = st.st_voc <- voc in
  (* let () = Voc.print_voc "vocname" voc in *)
  let () = read_to_end() in
  (close_state st; (first_tk, res_tk))

let tokenize filename mv voc_dir =
  let (tk, _) = read_tokens filename mv voc_dir in tk

let rec print_seq tk =
  print_token tk; match tk.tk_next with Some tk1 -> print_seq tk1 | None -> print_string("None\n")

let text_to_xml s =
  let buf = Buffer.create (String.length s) in
  let f ch = match ch with
        '&'  -> Buffer.add_string buf "&amp;"
      | '<'  -> Buffer.add_string buf "&lt;"
      | '>'  -> Buffer.add_string buf "&gt;"
      | '\'' -> Buffer.add_string buf "@apos;"
      | '"'  -> Buffer.add_string buf "&quot;" 
      | '\x00' .. '\x7F' -> Buffer.add_char buf ch
      | _ -> Buffer.add_string buf ("&#" ^ (string_of_int (Char.code ch)) ^ ";")
  in (String.iter f s; Buffer.contents buf)

(** Returns a pair (tag_name, attributes as association list) *)
let xml_tk_tag tk =
  let t = tk.tk_type in
  if t < Miz_comma then ("rw", [])  else
  if t < Miz_tk_numeral then ("Special-Symbol", []) else
  match t with
    Miz_tk_numeral -> ("Numeral", [])
  | Miz_tk_ident -> ("Identifier", [])
  | Miz_tk_symbol sq -> (match sq with
      Voc.Predicate -> ("Predicate-Symbol", [])
    | Voc.Functor i -> ("Functor-Symbol", [("priority", string_of_int i)])
    | Voc.Mode      -> ("Mode-Symbol", [])
    | Voc.Structure -> ("Structure-Symbol", [])
    | Voc.Selector  -> ("Selector-Symbol", [])
    | Voc.Attribute -> ("Attribute-Symbol", [])
    | Voc.Left_functor_bracket -> ("Left-Functor-Bracket", [])
    | Voc.Right_functor_bracket -> ("Right-Functor-Bracket", [])
    )
  | Miz_tk_whitespace -> ("Whitespace", [("text", tk.tk_text)])
  | Miz_tk_comment -> ("Comment", [])
  | _ -> ("", [])

let attr_to_str attr =
  let buf = Buffer.create 16 in
  let f buf p = match p with (name, value) -> (Buffer.add_string buf (" " ^ name ^ "=\"" ^ value ^ "\""); buf) in
  Buffer.contents (List.fold_left f buf attr)

let tk_to_xml tk =
  let tp = tk.tk_type in
  let text = tk.tk_text in
  let (tag_name, tag_attr) = xml_tk_tag tk in
  match tp with
    Miz_tk_eol -> "<eol/>\n"
  | Miz_tk_whitespace -> "<" ^ tag_name ^ (attr_to_str tag_attr) ^ " />"
  | Miz_tk_eof -> ""
  | _ -> "<" ^ tag_name ^ (attr_to_str tag_attr) ^ ">" ^ (text_to_xml text) ^ "</" ^ tag_name ^ ">"

let tk_seq_to_xml tk =
  let buf = Buffer.create 16 in
  let rec f tk_opt =
    match tk_opt with
    Some tk1 -> (Buffer.add_string buf (tk_to_xml tk1); f tk1.tk_next)
    | None -> ()
  in
    (f (Some tk); Buffer.contents buf)

let string_to_file s filename =
  let out = open_out filename in
  let () = output_string out s in
  close_out out

let miz_file_to_xml_file vct_map voc_dir miz_filename xml_filename =
  let (tk, _) = read_tokens miz_filename vct_map voc_dir in
  let s = "<Article>" ^ (tk_seq_to_xml tk) ^ "</Article>" in
  string_to_file s xml_filename

let miz_to_xml vct_map voc_dir filename =
  let (tk, _) = read_tokens filename vct_map voc_dir in
  let s = "<Article>" ^ (tk_seq_to_xml tk) ^ "</Article>" in
  let xml_filename = (Filename.chop_extension filename) ^ ".tokens.xml" in
  string_to_file s xml_filename

let miz_to_xml_e vct_map voc_dir filename =
  try
    miz_to_xml vct_map voc_dir filename
  with Token_error (line, pos) -> prerr_endline ("Token error at line " ^ (string_of_int line) ^ ", pos " ^ (string_of_int pos))
