(*
 * This program create a "defs.txt" file in the current directory.
 * This file contains for each vocabulary symbol the list of articles in
 * which this symbol is defined or redefined. Note that symonyms and antonyms
 * also are supported.
 *
 * Date: 2006-10-30
 * Copyright (C) 2006 Michael Nedzelsky (MichaelNedzelsky@yandex.ru)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Lexer
open Mizast

let mizar_dir = Sys.getenv("MIZFILES");;
let miz_article_path name = Filename.concat (Filename.concat mizar_dir "mml") (name ^ ".miz");;
let mml_vct = Filename.concat mizar_dir "mml.vct"
let mml_lar = Filename.concat mizar_dir "mml.lar"
let voc_dir = ref (Sys.getcwd())

let read_names filename =
  let i_s = open_in filename in
  let ls = ref [] in
  try
    while true do
      ls := (input_line i_s) :: !ls
    done;
    List.rev !ls
  with End_of_file -> (close_in i_s; List.rev !ls)

let names = read_names mml_lar
(* let names = ["abcmiz_0"; "xboole_0"; "xboole_1"; "xcmplx_0"] *)

let vct_map = Voc.read_vct_file mml_vct

module SetOfString = Set.Make(String)

module MapOfString = Map.Make(String)

let add_symbol_definition map symbol name =
  let (def_list, redef_list) = try MapOfString.find symbol map with Not_found -> ([], []) in
  MapOfString.add symbol (def_list @ [name], redef_list) map

let add_symbol_redefinition map symbol name =
  let (def_list, redef_list) = try MapOfString.find symbol map with Not_found -> ([], []) in
  MapOfString.add symbol (def_list, redef_list @ [name]) map

let add_tk_definition map prefix tk name =
  let symbol = prefix ^ tk.tk_text in
  let value = name ^ " at line " ^ (string_of_int tk.tk_line) in
  let (def_list, redef_list) = try MapOfString.find symbol map with Not_found -> ([], []) in
  MapOfString.add symbol (def_list @ [value], redef_list) map

let add_tk_redefinition map prefix tk name =
  let symbol = prefix ^ tk.tk_text in
  let value = name ^ " at line " ^ (string_of_int tk.tk_line) in
  let (def_list, redef_list) = try MapOfString.find symbol map with Not_found -> ([], []) in
  MapOfString.add symbol (def_list, redef_list @ [value]) map

(* Attribute-Pattern = Locus "is" Attribute-Symbol . *)
let get_tk_from_attribute_pattern node = (List.nth node.nt_sublist 1).nt_first_tk

(* Predicate-Pattern = [ Loci ] Predicate-Symbol [ Loci ] . *)
let get_tk_from_predicate_pattern node = let p = List.find (fun x -> x.nt_type = Predicate_symbol) node.nt_sublist in p.nt_first_tk

let get_prefix_tk_from_functor_pattern node =
  try
    let p = List.find (fun x -> x.nt_type = Functor_symbol) node.nt_sublist in ("O", p.nt_first_tk)
  with Not_found -> ("K", (List.hd node.nt_sublist).nt_first_tk)

(* Mode-Pattern = Mode-Symbol [ "of" Loci ] . *)
let get_tk_from_mode_pattern node = (List.hd node.nt_sublist).nt_first_tk

let get_prefix_tk_from_attr_or_pred node =
  match node.nt_type with
  | Attribute_pattern -> ("V", get_tk_from_attribute_pattern node)
  | Predicate_pattern -> ("R", get_tk_from_predicate_pattern node)
  | _ -> failwith "Unexpected node type in get_prefix_tk_from_attr_or_pred"

let get_prefix_tk_from_notation_declaration node =
 match node.nt_type with
 | Attribute_synonym -> get_prefix_tk_from_attr_or_pred (List.hd node.nt_sublist)
 | Attribute_antonym -> get_prefix_tk_from_attr_or_pred (List.hd node.nt_sublist)
 | Functor_synonym -> get_prefix_tk_from_functor_pattern (List.hd node.nt_sublist)
 | Mode_synonym -> ("M", get_tk_from_mode_pattern (List.hd node.nt_sublist))
 | Predicate_synonym -> ("R", get_tk_from_predicate_pattern (List.hd node.nt_sublist))
 | Predicate_antonym -> ("R", get_tk_from_predicate_pattern (List.hd node.nt_sublist))
 | _ -> failwith "Unexpected node type in get_prefix_tk_from_notation_declaration"

let add_notation_declaration map node name =
  let (prefix, tk) = get_prefix_tk_from_notation_declaration node in
  add_tk_definition map prefix tk name

let add_mode_definition map node name =
  let p = List.hd node.nt_sublist in
  add_tk_definition map "M" p.nt_first_tk name

(* Attribute-Definition = "attr" Attribute-Pattern "means" Definiens ";" Correctness-Conditions . *)
let add_attribute_definition map node name =
  let p = List.nth node.nt_sublist 0 in
  let tk = get_tk_from_attribute_pattern p in
  add_tk_definition map "V" tk name

(* Structure-Definition = "struct" [ "(" Ancestors ")" ] Structure-Symbol [ "over" Loci ] "(#" Fields "#)" ";" . *)
let add_structure_definition map node name =
  let p0 = List.nth node.nt_sublist 0 in
  let p = if p0.nt_type = Structure_symbol then p0 else List.nth node.nt_sublist 1 in
  add_tk_definition map "G" p.nt_first_tk name

let add_structure_redefinition map node name =  
  let p0 = List.nth node.nt_sublist 0 in
  let p = if p0.nt_type = Structure_symbol then p0 else List.nth node.nt_sublist 1 in
  add_tk_redefinition map "G" p.nt_first_tk name

(* Functor-Definition =  "func" Functor-Pattern [ Specification ] [ ( "means" | "equals" ) Definiens ] ";" Correctness-Conditions { Functor-Property } . *)
let add_functor_definition map node name =
  let p = List.nth node.nt_sublist 0 in
  let (prefix, tk) = get_prefix_tk_from_functor_pattern p in
  add_tk_definition map prefix tk name

let add_functor_redefinition map node name =
  let p = List.nth node.nt_sublist 0 in
  let (prefix, tk) = get_prefix_tk_from_functor_pattern p in
  add_tk_redefinition map prefix tk name

let add_predicate_definition map node name =
  let p0 = List.nth node.nt_sublist 0 in
  let p = List.find (fun x -> x.nt_type = Predicate_symbol) p0.nt_sublist in
  add_tk_definition map "R" p.nt_first_tk name

let add_predicate_redefinition map node name =
  let p0 = List.nth node.nt_sublist 0 in
  let p = List.find (fun x -> x.nt_type = Predicate_symbol) p0.nt_sublist in
  add_tk_redefinition map "R" p.nt_first_tk name

let add_mode_redefinition map node name =
  let p = List.hd node.nt_sublist in
  add_tk_redefinition map "M" p.nt_first_tk name

let add_attribute_redefinition map node name =
  let p = List.nth node.nt_sublist 0 in
  let tk = get_tk_from_attribute_pattern p in
  add_tk_redefinition map "V" tk name

let add_definition map node name =
  match node.nt_type with
  | Mode_definition -> add_mode_definition map node name
  | Attribute_definition -> add_attribute_definition map node name
  | Functor_definition -> add_functor_definition map node name
  | Structure_definition -> add_structure_definition map node name
  | Predicate_definition -> add_predicate_definition map node name
  | _ -> map

let add_redefinition map node name =
  match node.nt_type with
  | Mode_definition -> add_mode_redefinition map node name
  | Attribute_definition -> add_attribute_redefinition map node name
  | Functor_definition -> add_functor_redefinition map node name
  | Structure_definition -> add_structure_redefinition map node name
  | Predicate_definition -> add_predicate_redefinition map node name
  | _ -> map

let add_redefinition_block map node_list name =
  let add_node map node =
    match node.nt_type with
    | Definition -> add_redefinition map (List.hd node.nt_sublist) name
    | Definition_item -> map
    | _ -> map
  in
  let rec add_list map list =
    match list with
    | [] -> map
    | h :: t -> add_list (add_node map h) t
  in
  let result = add_list map node_list in
  (* let () = print_endline (if MapOfString.is_empty result then "1 empty" else "1 non-empty") in *)
  result

(* Definitional-Block = "definition" { ( Definition-Item | Definition ) } { Redefinition-Block } "end" . *)
let add_definitional_block map node_list name =
  let add_node map node =
    match node.nt_type with
    | Definition -> add_definition map (List.hd node.nt_sublist) name
    | Definition_item -> map
    | Redefinition_block -> add_redefinition_block map node.nt_sublist name
    | _ -> map
  in
  let rec add_list map list =
    match list with
    | [] -> map
    | h :: t -> add_list (add_node map h) t
  in
  let result = add_list map node_list in
  (* let () = print_endline (if MapOfString.is_empty result then "1 empty" else "1 non-empty") in *)
  result

let rec add_to_map name map node =
  match node.nt_type with
  | Definitional_block -> add_definitional_block map node.nt_sublist name
  | Notation_declaration -> add_notation_declaration map (List.hd node.nt_sublist) name
  | _ -> List.fold_left (add_to_map name) map node.nt_sublist

let do_one map name =
  let miz_filename = miz_article_path name in
  begin
    print_string name;
    print_newline();
    try
      let node = Mizast.miz_file_to_ast vct_map !voc_dir miz_filename in
      let result = add_to_map name map node in
      (* let () = print_endline (if MapOfString.is_empty result then "empty" else "non-empty") in *)
      result
    with Lexer.Token_error (line, pos) ->
      (prerr_endline (name ^ ": Token error at line " ^ (string_of_int line) ^ ", pos " ^ (string_of_int pos)); map)
  end

let g os key (def_list, redef_list) =
  let () = output_string os (key ^ "\n") in
  let () = output_string os "  defined in:\n" in
  let () = List.iter (fun x -> output_string os ("    " ^ x ^ "\n")) def_list in
  let () = output_string os "  redefined in:\n" in
  let () = List.iter (fun x -> output_string os ("    " ^ x ^ "\n")) redef_list in
  let () = output_string os "\n" in
  ()

let f () =
  let map = List.fold_left do_one MapOfString.empty names in
  let os = open_out "defs.txt" in
  let () = MapOfString.iter (g os) map in
  close_out os

let () = f ()
