(*
 * Usefull functions for work with Mizar vocabularies.
 * Date: 2006-10-15
 * Copyright (C) 2006 Michael Nedzelsky (MichaelNedzelsky@yandex.ru)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

let is_ws ch = match ch with '\t' | ' ' -> true | _ -> false

(* Mizar vocabulary *)
type symbol_qualifier = Predicate | Functor of int | Mode | Structure
  | Selector | Attribute | Left_functor_bracket | Right_functor_bracket

exception Voc_symbol_error of string;;

(** Returns the largest pos1 such that all the chars s[pos], s[pos+1], ... , s[pos1-1] are digits. *)
let rec last_digit_pos s pos = if pos >= String.length s then pos else
  match s.[pos] with
  '0' .. '9' -> last_digit_pos s (pos+1)
  | _ -> pos

(* Result is (Some number or None, pos_of_first_char_after_number). *)
let read_short_int s pos = let pos1 = last_digit_pos s pos in
  let len = pos1 - pos in
  if len = 0 then (None, pos1) else
  if len > 3 then failwith "Number must be in range 0..255" else
  if len>0 && s.[pos] = '0' then failwith "Number which is greater than 0 must be started with a positive digit" else
  let n = int_of_string (String.sub s pos len) in
  if n > 255 then failwith "Number must be in range 0..255" else (Some n, pos1)

(* Returns the least new_pos >= pos such that new_pos >= String.length s or s.[new_pos] is whitespace. *)
let rec first_ws_or_eol s pos =
  let len = String.length s in
    if pos < len then
      let ch = s.[pos] in
        if is_ws ch then pos else first_ws_or_eol s (pos+1)
    else pos

(* Returns the least new_pos >= pos such that new_pos >= String.length s or s.[new_pos] is not a whitespace. *)
let rec first_non_ws s pos =
  let len = String.length s in
    if pos < len then
      let ch = s.[pos] in
        if is_ws ch then first_non_ws s (pos+1) else pos
    else pos

(* Reads one symbol from the vocabulary line. *)
let read_one_symbol s =
  let read_opt_priority pos =
    let pos1 = first_non_ws s pos in
    let (opt_n, new_pos)  = read_short_int s pos1 in
    let n = match opt_n with Some x -> x | None -> 64 in  (n, new_pos)
  in
  let test_eol pos = if pos < String.length s then raise (Voc_symbol_error ("expected end of line at pos " ^ (string_of_int pos)))
  in
  let ch = s.[0] in
  let pos = first_ws_or_eol s 1 in
  let str = String.sub s 1 (pos-1) in
  let res =
    if ch = 'O' then let (n, pos) = read_opt_priority pos in (test_eol pos); Functor n
    else begin
      test_eol pos;
      match ch with
	'R' -> Predicate
      | 'M' -> Mode
      | 'G' -> Structure
      | 'U' -> Selector
      | 'V' -> Attribute
      | 'K' -> Left_functor_bracket
      | 'L' -> Right_functor_bracket
      | _ -> raise (Voc_symbol_error "first char must be a qualifier.")
    end
  in (res, str)

module MapOfString = Map.Make(String)

let print_voc voc_name m =
  let () = print_string (voc_name ^ "\n") in
  MapOfString.iter (fun key _ -> print_string ("  " ^ key ^ "\n")) m

let print_vocs mv =
  MapOfString.iter print_voc mv

type t_vocs = symbol_qualifier MapOfString.t MapOfString.t
type t_voc = symbol_qualifier MapOfString.t

(* Returns map. *)
let rec read_voc map i_s =
  try
    let s = input_line i_s in
    let entry = read_one_symbol s in
    read_voc (MapOfString.add (snd entry) (fst entry) map) i_s
  with End_of_file -> map

let read_voc_file filename =
  let i_s = open_in filename in
  let m = read_voc MapOfString.empty i_s in
  let () = close_in i_s in
  m

let find_voc_file voc_dir voc_name =
  let s = String.lowercase (voc_name ^ ".voc") in
  let dir_a = Sys.readdir voc_dir in
  let len = Array.length dir_a in
  let rec find i =
  if i >= len then raise Not_found else
    let item = dir_a.(i) in
    if (String.lowercase item) = s then item else find (i+1) in
  find 0

let read_vocabulary voc_dir voc_name = read_voc_file (Filename.concat voc_dir (find_voc_file voc_dir voc_name))

let rec read_vocs mv s i_s =
  (* let () = print_string s in *)
  let rec read_voc m =
    try
      let s = input_line i_s in
      if (String.length s > 0) & (s.[0] = '#') then (m, s) else
        let entry = read_one_symbol s in
        let (m1,s1) = read_voc (MapOfString.add (snd entry) (fst entry) m) in
        (m1, s1)
    with End_of_file -> (m, "")
  in
  if s = "" then (mv, s) else
    let voc_name = String.sub s 1 (String.length s - 1) in
    (* let () = print_string ("voc_name = " ^ voc_name ^ "\n") in *)
    let () = ignore (input_line i_s) in
    let (m2, s2) = read_voc MapOfString.empty in   
    (* let () = print_string ("s2 = " ^ s2 ^ "\n") in *)
    let mv1 = MapOfString.add voc_name m2 mv in
    if s2 = "" then (mv1, "") else read_vocs mv1 s2 i_s
  
(** Returns the map of vocabularies. *)
let read_vct_file filename =
  let i_s = open_in filename in
  let s = input_line i_s in
  let (mv, _) = read_vocs MapOfString.empty s i_s in
  let () = close_in i_s in
  mv

